#!/usr/bin/env python
import struct
import threading
from array import array
from collections import deque
from typing import Optional

import usb.backend.libusb0 as libusb0
import usb.backend.libusb1 as libusb1
import usb.backend.openusb as openusb
import usb.core
import usb.util
from usb.core import USBError, USBTimeoutError

from pyxcp.transport.base import BaseTransport
from pyxcp.utils import short_sleep


RECV_SIZE = 16384
FIVE_MS = 5_000_000  # Five milliseconds in nanoseconds.


class Usb(BaseTransport):
    """"""

    HEADER = struct.Struct("<2H")
    HEADER_SIZE = HEADER.size

    def __init__(self, config=None, policy=None, transport_layer_interface: Optional[usb.core.Device] = None):
        super().__init__(config, policy, transport_layer_interface)
        self.load_config(config)
        self.serial_number: str = self.config.serial_number
        self.vendor_id: int = self.config.vendor_id
        self.product_id: int = self.config.product_id
        self.configuration_number: int = self.config.configuration_number
        self.interface_number: int = self.config.interface_number
        self.library: str = self.config.library
        self.header_format: str = self.config.header_format
        self.library = self.config.get("library")
        self.device = None

        ## IN-EP (RES/ERR, DAQ, and EV/SERV) Parameters.
        self.in_ep_number: int = self.config.in_ep_number
        self.in_ep_transfer_type = self.config.in_ep_transfer_type
        self.in_ep_max_packet_size: int = self.config.in_ep_max_packet_size
        self.in_ep_polling_interval: int = self.config.in_ep_polling_interval
        self.in_ep_message_packing = self.config.in_ep_message_packing
        self.in_ep_alignment = self.config.in_ep_alignment
        self.in_ep_recommended_host_bufsize: int = self.config.in_ep_recommended_host_bufsize

        ## OUT-EP (CMD and STIM) Parameters.
        self.out_ep_number: int = self.config.out_ep_number

        self.device: Optional[usb.core.Device] = None
        self.status = 0

        self._packet_listener = threading.Thread(
            target=self._packet_listen,
            args=(),
            kwargs={},
        )
        self._packets = deque()

    def connect(self):
        if self.library:
            for backend_provider in (libusb1, libusb0, openusb):
                backend = backend_provider.get_backend(find_library=lambda x: self.library)
                if backend:
                    break
        else:
            backend = None

        if self.vendor_id and self.product_id:
            kwargs = {
                "find_all": True,
                "idVendor": self.vendor_id,
                "idProduct": self.product_id,
                "backend": backend,
            }
        else:
            kwargs = {
                "find_all": True,
                "backend": backend,
            }

        for device in usb.core.find(**kwargs):
            try:
                if device.serial_number.strip().strip("\0").lower() == self.serial_number.lower():
                    self.device = device
                    break
            except (USBError, USBTimeoutError):
                continue
        else:
            raise Exception(f"XCPonUSB - device with serial {self.serial_number!r} not found")

        current_configuration = self.device.get_active_configuration()
        if current_configuration.bConfigurationValue != self.configuration_number:
            self.device.set_configuration(self.configuration_number)
        cfg = self.device.get_active_configuration()

        interface = cfg[(self.interface_number, 0)]

        self.out_ep = interface[self.out_ep_number]
        self.in_ep = interface[self.in_ep_number]

        self.start_listener()
        self.status = 1  # connected

    def start_listener(self):
        super().start_listener()
        if self._packet_listener.is_alive():
            self._packet_listener.join()
        self._packet_listener = threading.Thread(target=self._packet_listen)
        self._packet_listener.start()

    def close(self):
        """Close the transport-layer connection and event-loop."""
        self.finish_listener()
        if self.listener.is_alive():
            self.listener.join()
        if self._packet_listener.is_alive():
            self._packet_listener.join()
        self.close_connection()

    def _packet_listen(self):
        close_event_set = self.closeEvent.is_set
        _packets = self._packets
        read = self.in_ep.read
        buffer = array("B", bytes(RECV_SIZE))
        buffer_view = memoryview(buffer)
        while True:
            try:
                if close_event_set():
                    return
                try:
                    recv_timestamp = self.timestamp.value
                    read_count = read(buffer, 100)  # 100ms timeout
                    if read_count != RECV_SIZE:
                        _packets.append((buffer_view[:read_count].tobytes(), recv_timestamp))
                    else:
                        _packets.append((buffer.tobytes(), recv_timestamp))
                except (USBError, USBTimeoutError):
                    # print(format_exc())
                    short_sleep()
                    continue
            except BaseException:  # noqa: B036
                # Note: catch-all only permitted if the intention is re-raising.
                self.status = 0  # disconnected
                break

    def listen(self):
        HEADER_UNPACK_FROM = self.HEADER.unpack_from
        HEADER_SIZE = self.HEADER_SIZE
        popleft = self._packets.popleft
        process_response = self.process_response
        close_event_set = self.closeEvent.is_set
        _packets = self._packets
        length: Optional[int] = None
        counter: int = 0
        data: bytearray = bytearray(b"")
        last_sleep: int = self.timestamp.value

        while True:
            if close_event_set():
                return
            count: int = len(_packets)
            if not count:
                short_sleep()
                last_sleep = self.timestamp.value
                continue
            for _ in range(count):
                bts, timestamp = popleft()
                data += bts
                current_size: int = len(data)
                current_position: int = 0
                while True:
                    if self.timestamp.value - last_sleep >= FIVE_MS:
                        short_sleep()
                        last_sleep = self.timestamp.value
                    if length is None:
                        if current_size >= HEADER_SIZE:
                            length, counter = HEADER_UNPACK_FROM(data, current_position)
                            current_position += HEADER_SIZE
                            current_size -= HEADER_SIZE
                        else:
                            data = data[current_position:]
                            break
                    else:
                        if current_size >= length:
                            response = data[current_position : current_position + length]
                            process_response(response, length, counter, timestamp)
                            current_size -= length
                            current_position += length
                            length = None
                        else:
                            data = data[current_position:]
                            break

    def send(self, frame):
        self.pre_send_timestamp = self.timestamp.value
        try:
            self.out_ep.write(frame)
        except (USBError, USBTimeoutError):
            # sometimes usb.core.USBError: [Errno 5] Input/Output Error is raised
            # even though the command is send and a reply is received from the device.
            # Ignore this here since a Timeout error will be raised anyway if
            # the device does not respond
            pass
        self.post_send_timestamp = self.timestamp.value

    def close_connection(self):
        if self.device is not None:
            usb.util.dispose_resources(self.device)



#!/usr/bin/env python

import can


def main():
    can.log.setLevel("ERROR")
    interfaces = can.detect_available_configs()
    print("-" * 80)
    if not interfaces:
        print("No CAN-interfaces installed on your system.")
    else:
        print("\nInstalled CAN-interfaces on your system:\n")
        interfaces = sorted(interfaces, key=lambda e: e["interface"])
        for interface in interfaces:
            print("\t{:20s} -- CHANNEL: {}".format(interface["interface"], interface["channel"]))


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import multiprocessing as mp
import os
import platform
import re
import subprocess  # nosec
import sys
import sysconfig
from pathlib import Path
from tempfile import TemporaryDirectory


TOP_DIR = Path(__file__).parent

print("Platform", platform.system())
uname = platform.uname()
if uname.system == "Darwin":
    os.environ["MACOSX_DEPLOYMENT_TARGET"] = "10.13"

VARS = sysconfig.get_config_vars()


def get_python_base() -> str:
    # Applies in this form only to Windows.
    if "base" in VARS and VARS["base"]:
        return VARS["base"]
    if "installed_base" in VARS and VARS["installed_base"]:
        return VARS["installed_base"]


def alternate_libdir(pth: str):
    base = Path(pth).parent
    libdir = Path(base) / "libs"
    if libdir.exists():
        # available_libs = os.listdir(libdir)
        return str(libdir)
    else:
        return ""


def get_py_config() -> dict:
    pynd = VARS["py_version_nodot"]  # Should always be present.
    include = sysconfig.get_path("include")  # Seems to be cross-platform.
    if uname.system == "Windows":
        base = get_python_base()
        library = f"python{pynd}.lib"
        libdir = Path(base) / "libs"
        if libdir.exists():
            available_libs = os.listdir(libdir)
            if library in available_libs:
                libdir = str(libdir)
            else:
                libdir = ""
        else:
            libdir = alternate_libdir(include)
    else:
        library = VARS["LIBRARY"]
        DIR_VARS = ("LIBDIR", "BINLIBDEST", "DESTLIB", "LIBDEST", "MACHDESTLIB", "DESTSHARED", "LIBPL")
        arch = None
        if uname.system == "Linux":
            arch = VARS.get("MULTIARCH", "")
        found = False
        for dir_var in DIR_VARS:
            if found:
                break
            dir_name = VARS.get(dir_var)
            if not dir_name:
                continue
            if uname.system == "Darwin":
                full_path = [Path(dir_name) / library]
            elif uname.system == "Linux":
                full_path = [Path(dir_name) / arch / library, Path(dir_name) / library]
            else:
                print("PF?", uname.system)
            for fp in full_path:
                print(f"Trying '{fp}'")
                if fp.exists():
                    print(f"found Python library: '{fp}'")
                    libdir = str(fp.parent)
                    found = True
                    break
        if not found:
            print("Could NOT locate Python library.")
            return dict(exe=sys.executable, include=include, libdir="", library=library)
    return dict(exe=sys.executable, include=include, libdir=libdir, library=library)


def banner(msg: str) -> None:
    print("=" * 80)
    print(str.center(msg, 80))
    print("=" * 80)


def build_extension(debug: bool = False, use_temp_dir: bool = False) -> None:
    print("build_ext::build_extension()")

    debug = bool(os.environ.get("DEBUG", 0)) or debug
    cfg = "Debug" if debug else "Release"
    print(f" BUILD-TYPE: {cfg!r}")

    py_cfg = get_py_config()

    cmake_args = [
        f"-DPython3_EXECUTABLE={py_cfg['exe']}",
        f"-DPython3_INCLUDE_DIR={py_cfg['include']}",
        f"-DCMAKE_BUILD_TYPE={cfg}",  # not used on MSVC, but no harm
    ]
    if py_cfg["libdir"]:
        cmake_args.append(f"-DPython3_LIBRARY={str(Path(py_cfg['libdir']) / Path(py_cfg['library']))}")

    build_args = ["--config Release", "--verbose"]
    # cmake -DCMAKE_EXPORT_COMPILE_COMMANDS=1 /path/to/src

    if sys.platform.startswith("darwin"):
        # Cross-compile support for macOS - respect ARCHFLAGS if set
        archs = re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", ""))
        if archs:
            cmake_args += ["-DCMAKE_OSX_ARCHITECTURES={}".format(";".join(archs))]

    if use_temp_dir:
        build_temp = Path(TemporaryDirectory(suffix=".build-temp").name) / "extension_it_in"
    else:
        build_temp = Path(".")
    # print("cwd:", os.getcwd(), "build-dir:", build_temp, "top:", str(TOP_DIR))
    if not build_temp.exists():
        build_temp.mkdir(parents=True)

    banner("Step #1: Configure")
    # cmake_args += ["--debug-output"]
    subprocess.run(["cmake", "-S", str(TOP_DIR), *cmake_args], cwd=build_temp, check=True)  # nosec

    cmake_args += [f"--parallel {mp.cpu_count()}"]

    banner("Step #2: Build")
    # build_args += ["-DCMAKE_VERBOSE_MAKEFILE:BOOL=ON"]
    subprocess.run(["cmake", "--build", str(build_temp), *build_args], cwd=TOP_DIR, check=True)  # nosec

    banner("Step #3: Install")
    subprocess.run(["cmake", "--install", "."], cwd=build_temp, check=True)  # nosec
    subprocess.run(["cmake", "--install", build_temp], cwd=TOP_DIR, check=True)  # nosec


if __name__ == "__main__":
    includes = subprocess.getoutput("pybind11-config --cmakedir")  # nosec
    os.environ["pybind11_DIR"] = includes
    build_extension(False)



import pytest

from pyxcp.daq_stim.optimize import McObject, make_continuous_blocks
from pyxcp.daq_stim.optimize.binpacking import Bin, first_fit_decreasing


@pytest.fixture
def blocks():
    return [
        McObject(name="", address=0x000E10BA, length=2),
        McObject(name="", address=0x000E10BE, length=2),
        McObject(name="", address=0x000E41F4, length=4),
        McObject(name="", address=0x000E51FC, length=4),
        McObject(name="", address=0x00125288, length=4),
        McObject(name="", address=0x00125294, length=4),
        McObject(name="", address=0x001252A1, length=1),
        McObject(name="", address=0x001252A4, length=4),
        McObject(name="", address=0x00125438, length=3),
        McObject(name="", address=0x0012543C, length=1),
    ]


def test_pack_to_single_bin(blocks):
    BIN_SIZE = 253
    bins = first_fit_decreasing(items=blocks, bin_size=BIN_SIZE)

    assert len(bins) == 1
    bin0 = bins[0]
    assert bin0.residual_capacity == BIN_SIZE - 29
    assert bin0.entries == [
        McObject(name="", address=0x000E41F4, length=4),
        McObject(name="", address=0x000E51FC, length=4),
        McObject(name="", address=0x00125288, length=4),
        McObject(name="", address=0x00125294, length=4),
        McObject(name="", address=0x001252A4, length=4),
        McObject(name="", address=0x00125438, length=3),
        McObject(name="", address=0x000E10BA, length=2),
        McObject(name="", address=0x000E10BE, length=2),
        McObject(name="", address=0x001252A1, length=1),
        McObject(name="", address=0x0012543C, length=1),
    ]


def test_pack_empty_block_set():
    BIN_SIZE = 253
    bins = first_fit_decreasing(items=[], bin_size=BIN_SIZE)
    assert bins == [Bin(size=BIN_SIZE)]


def test_pack_to_multiple_bins1(blocks):
    BIN_SIZE = 6
    bins = first_fit_decreasing(items=blocks, bin_size=BIN_SIZE)
    assert len(bins) == 6
    bin0, bin1, bin2, bin3, bin4, bin5 = bins
    assert bin0.residual_capacity == 0
    assert bin0.entries == [
        McObject(name="", address=0x000E41F4, length=4),
        McObject(name="", address=0x000E10BA, length=2),
    ]
    assert bin1.residual_capacity == 0
    assert bin1.entries == [
        McObject(name="", address=0x000E51FC, length=4),
        McObject(name="", address=0x000E10BE, length=2),
    ]
    assert bin2.residual_capacity == 0
    assert bin2.entries == [
        McObject(name="", address=0x00125288, length=4),
        McObject(name="", address=0x001252A1, length=1),
        McObject(name="", address=0x0012543C, length=1),
    ]
    assert bin3.residual_capacity == 2
    assert bin3.entries == [McObject(name="", address=0x00125294, length=4)]
    assert bin4.residual_capacity == 2
    assert bin4.entries == [McObject(name="", address=0x001252A4, length=4)]
    assert bin5.residual_capacity == 3
    assert bin5.entries == [McObject(name="", address=0x00125438, length=3)]


def test_binpacking_raises(blocks):
    BIN_SIZE = 7
    with pytest.raises(ValueError):
        first_fit_decreasing(items=[McObject(name="", address=0x1000, length=32)], bin_size=BIN_SIZE)


def test_binpacking_works(blocks):
    BIN_SIZE = 7
    first_fit_decreasing(items=[McObject(name="", address=0x1000, length=7)], bin_size=BIN_SIZE)


def test_make_continuous_blocks1():
    BLOCKS = [
        McObject(name="", address=0x000E0002, length=2),
        McObject(name="", address=0x000E0008, ext=23, length=4),
        McObject(name="", address=0x000E0004, length=4),
        McObject(name="", address=0x000E000C, ext=23, length=4),
        McObject(name="", address=0x000E0000, length=2),
    ]
    bins = make_continuous_blocks(chunks=BLOCKS)
    assert bins == [
        McObject(
            name="",
            address=917504,
            ext=0,
            length=8,
            components=[
                McObject(name="", address=917504, ext=0, length=2, components=[]),
                McObject(name="", address=917506, ext=0, length=2, components=[]),
                McObject(name="", address=917508, ext=0, length=4, components=[]),
            ],
        ),
        McObject(
            name="",
            address=917512,
            ext=23,
            length=8,
            components=[
                McObject(name="", address=917512, ext=23, length=4, components=[]),
                McObject(name="", address=917516, ext=23, length=4, components=[]),
            ],
        ),
    ]


def test_make_continuous_blocks2():
    BLOCKS = [
        McObject(name="", address=0x000E0002, length=2),
        McObject(name="", address=0x000E0008, length=4),
        McObject(name="", address=0x000E0004, length=4),
        McObject(name="", address=0x000E000C, length=4),
        McObject(name="", address=0x000E0000, length=2),
    ]
    bins = make_continuous_blocks(chunks=BLOCKS)
    assert bins == [
        McObject(
            name="",
            address=917504,
            ext=0,
            length=16,
            components=[
                McObject(name="", address=917504, ext=0, length=2, components=[]),
                McObject(name="", address=917506, ext=0, length=2, components=[]),
                McObject(name="", address=917508, ext=0, length=4, components=[]),
                McObject(name="", address=917512, ext=0, length=4, components=[]),
                McObject(name="", address=917516, ext=0, length=4, components=[]),
            ],
        )
    ]


def test_make_continuous_blocks3():
    BLOCKS = [
        McObject(name="", address=0x000E0002, ext=0x01, length=2),
        McObject(name="", address=0x000E0008, ext=0x03, length=4),
        McObject(name="", address=0x000E0004, ext=0x02, length=4),
        McObject(name="", address=0x000E000C, ext=0x04, length=4),
        McObject(name="", address=0x000E0000, ext=0x00, length=2),
    ]
    bins = make_continuous_blocks(chunks=BLOCKS)
    assert bins == [
        McObject(
            name="", address=917504, ext=0, length=2, components=[McObject(name="", address=917504, ext=0, length=2, components=[])]
        ),
        McObject(
            name="", address=917506, ext=1, length=2, components=[McObject(name="", address=917506, ext=1, length=2, components=[])]
        ),
        McObject(
            name="", address=917508, ext=2, length=4, components=[McObject(name="", address=917508, ext=2, length=4, components=[])]
        ),
        McObject(
            name="", address=917512, ext=3, length=4, components=[McObject(name="", address=917512, ext=3, length=4, components=[])]
        ),
        McObject(
            name="", address=917516, ext=4, length=4, components=[McObject(name="", address=917516, ext=4, length=4, components=[])]
        ),
    ]


def test_mc_object_len_zero():
    with pytest.raises(ValueError):
        McObject(name="", address=0, ext=0, length=0)


def test_mc_object_ok():
    McObject(name="", address=0, ext=0, length=1)



#!/usr/bin/env python

MAP_NAMES = {
    1: "BorlandC 16 Bit",
    2: "M166",
    3: "Watcom",
    4: "HiTech HC05",
    6: "IEEE",
    7: "Cosmic",
    8: "SDS",
    9: "Fujitsu Softune 1(.mp1)",
    10: "GNU",
    11: "Keil 16x",
    12: "BorlandC 32 Bit",
    13: "Keil 16x (static)",
    14: "Keil 8051",
    15: "ISI",
    16: "Hiware HC12",
    17: "TI TMS470",
    18: "Archimedes",
    19: "COFF",
    20: "IAR",
    21: "VisualDSP",
    22: "GNU 16x",
    23: "GNU VxWorks",
    24: "GNU 68k",
    25: "DiabData",
    26: "VisualDSP DOS",
    27: "HEW SH7055",
    28: "Metrowerks",
    29: "Microsoft standard",
    30: "ELF/DWARF 16 Bit",
    31: "ELF/DWARF 32 Bit",
    32: "Fujitsu Softune 3..8(.mps)",
    33: "Microware Hawk",
    34: "TI C6711",
    35: "Hitachi H8S",
    36: "IAR HC12",
    37: "Greenhill Multi 2000",
    38: "LN308(MITSUBISHI) for M16C/80",
    39: "COFF settings auto detected",
    40: "NEC CC78K/0 v35",
    41: "Microsoft extended",
    42: "ICCAVR",
    43: "Omf96 (.m96)",
    44: "COFF/DWARF",
    45: "OMF96 Binary (Tasking C196)",
    46: "OMF166 Binary (Keil C166)",
    47: "Microware Hawk Plug&Play ASCII",
    48: "UBROF Binary (IAR)",
    49: "Renesas M32R/M32192 ASCII",
    50: "OMF251 Binary (Keil C251)",
    51: "Microsoft standard VC8",
    52: "Microsoft VC8 Release Build (MATLAB DLL)",
    53: "Microsoft VC8 Debug Build (MATLAB DLL)",
    54: "Microsoft VC8 Debug file (pdb)",
}

"""
3.0  Automatic detection sequence

-  The master (CANape) sends the command GET_ID with the idType = 219 (0xDB hex) to the ECU
-  The ECU sets the MTA (memory transfer address) pointer to the first byte of the data block containing the
MAP file identification and returns the number of bytes which need to be uploaded to the master as well as
the counter which defines the appropriate memory segment
-  The master requests the specified number of bytes from the ECU by sending UPLOAD commands
-  The master looks for the appropriate MAP file in the MAP file directory and updates the addresses in the
a2l file accordingly

Command sequence GET_ID

Id of the MAP format

Counter

MAP name

"""


def mapfile_name(name, counter, fmt):
    return f"{fmt:2d}{counter:d}{name:s}.map"



#!/usr/bin/env python
"""Bin-packing algorithms.
"""
from typing import List, Optional

from pyxcp.cpp_ext import Bin


def first_fit_decreasing(items, bin_size: int, initial_bin_size: Optional[int] = None) -> List[Bin]:
    """bin-packing with first-fit-decreasing algorithm.

    Parameters
    ----------
    items: list
        items that need to be stored/allocated.

    bin_size: int

    Returns
    -------
    list
        Resulting bins
    """
    if initial_bin_size is None:
        initial_bin_size = bin_size
    # bin_size = max(bin_size, initial_bin_size)
    bins = [Bin(size=initial_bin_size)]  # Initial bin
    for item in sorted(items, key=lambda x: x.length, reverse=True):
        if item.length > bin_size:
            raise ValueError(f"Item {item!r} is too large to fit in a {bin_size} byte sized bin.")
        for bin in bins:
            if bin.residual_capacity >= item.length:
                bin.append(item)
                bin.residual_capacity -= item.length
                break
        else:
            new_bin = Bin(size=bin_size)
            bins.append(new_bin)
            new_bin.append(item)
            new_bin.residual_capacity -= item.length
    return bins



#!/usr/bin/env python
"""Demostrates how to use frame recording policies including recorder extension.
"""
from pprint import pprint

from pyxcp.cmdline import ArgumentParser
from pyxcp.transport.base import FrameRecorderPolicy, StdoutPolicy  # noqa: F401


ap = ArgumentParser(description="pyXCP frame recording policy example.")

LOG_FILE = "pyxcp"

policy = FrameRecorderPolicy(LOG_FILE)
use_recorder = True

# policy = StdoutPolicy()  # You may also try this one.
# use_recorder = False

with ap.run(policy=policy) as x:
    x.connect()
    if x.slaveProperties.optionalCommMode:
        x.getCommModeInfo()
    identifier = x.identifier(0x01)
    print("\nSlave Properties:")
    print("=================")
    print(f"ID: {identifier!r}")
    pprint(x.slaveProperties)
    x.disconnect()

if use_recorder:
    from pyxcp.recorder import XcpLogFileReader
    from pyxcp.utils import hexDump

    try:
        import pandas  # noqa: F401
    except ImportError:
        has_pandas = False
    else:
        has_pandas = True

    reader = XcpLogFileReader(LOG_FILE)
    hdr = reader.get_header()  # Get file information.
    print("\nRecording file header")
    print("=====================\n")
    print(hdr)
    print("\nRecorded frames")
    print("===============\n")
    print("CAT         CTR  TS             PAYLOAD")
    print("-" * 80)
    for category, counter, timestamp, payload in reader:
        print(f"{category.name:8} {counter:6}  {timestamp:7.7f} {hexDump(payload)}")
    print("-" * 80)
    reader.reset_iter()  # reader acts as an Python iterator -- can be reseted with this non-standard method.
    if has_pandas:
        print("\nRecordings as Pandas stuff")
        print("==========================\n")
        df = reader.as_dataframe()  # Return recordings as Pandas DataFrame.
        print(df.info())
        print(df.head(60))



import pytest

from pyxcp.asam import types


def testEncodeUint32_0():
    assert types.A_Uint32("<").encode(3415750566) == b"\xa67\x98\xcb"


def testDecodeUint32_0():
    assert types.A_Uint32("<").decode((0xA6, 0x37, 0x98, 0xCB)) == 3415750566


def testLittleEndian():
    assert isinstance(types.AsamBaseType(types.INTEL), types.AsamBaseType)


def testBigEndian():
    assert isinstance(types.AsamBaseType(types.MOTOROLA), types.AsamBaseType)


def testInvalidByteOrderRaisesTypeError():
    with pytest.raises(ValueError):
        types.AsamBaseType("#")



#!/usr/bin/env python

# from pprint import pprint
from time import time_ns
from typing import Dict, List, TextIO

from pyxcp import types
from pyxcp.config import get_application
from pyxcp.cpp_ext import DaqList
from pyxcp.daq_stim.optimize import make_continuous_blocks
from pyxcp.daq_stim.optimize.binpacking import first_fit_decreasing
from pyxcp.recorder import DaqOnlinePolicy as _DaqOnlinePolicy
from pyxcp.recorder import DaqRecorderPolicy as _DaqRecorderPolicy
from pyxcp.recorder import MeasurementParameters
from pyxcp.utils import CurrentDatetime


DAQ_ID_FIELD_SIZE = {
    "IDF_ABS_ODT_NUMBER": 1,
    "IDF_REL_ODT_NUMBER_ABS_DAQ_LIST_NUMBER_BYTE": 2,
    "IDF_REL_ODT_NUMBER_ABS_DAQ_LIST_NUMBER_WORD": 3,
    "IDF_REL_ODT_NUMBER_ABS_DAQ_LIST_NUMBER_WORD_ALIGNED": 4,
}

DAQ_TIMESTAMP_SIZE = {
    "S1": 1,
    "S2": 2,
    "S4": 4,
}


class DaqProcessor:
    def __init__(self, daq_lists: List[DaqList]):
        # super().__init__()
        self.daq_lists = daq_lists
        self.log = get_application().log

    def setup(self, start_datetime: CurrentDatetime | None = None, write_multiple: bool = True):
        self.daq_info = self.xcp_master.getDaqInfo()
        if start_datetime is None:
            start_datetime = CurrentDatetime(time_ns())
        self.start_datetime = start_datetime
        # print(self.start_datetime)
        try:
            processor = self.daq_info.get("processor")
            properties = processor.get("properties")
            resolution = self.daq_info.get("resolution")
            if properties["configType"] == "STATIC":
                raise TypeError("DAQ configuration is static, cannot proceed.")
            self.supports_timestampes = properties["timestampSupported"]
            self.supports_prescaler = properties["prescalerSupported"]
            self.supports_pid_off = properties["pidOffSupported"]
            if self.supports_timestampes:
                mode = resolution.get("timestampMode")
                self.ts_fixed = mode.get("fixed")
                self.ts_size = DAQ_TIMESTAMP_SIZE[mode.get("size")]
                ts_factor = types.DAQ_TIMESTAMP_UNIT_TO_NS[mode.get("unit")]
                ts_ticks = resolution.get("timestampTicks")
                self.ts_scale_factor = ts_factor * ts_ticks
            else:
                self.ts_size = 0
                self.ts_fixed = False
                self.ts_scale_factor = 0.0
            key_byte = processor.get("keyByte")
            header_len = DAQ_ID_FIELD_SIZE[key_byte["identificationField"]]
            max_dto = self.xcp_master.slaveProperties.maxDto
            self.min_daq = processor.get("minDaq")
            max_odt_entry_size = resolution.get("maxOdtEntrySizeDaq")
            max_payload_size = min(max_odt_entry_size, max_dto - header_len)
            # First ODT may contain timestamp.
            self.selectable_timestamps = False
            if not self.supports_timestampes:
                max_payload_size_first = max_payload_size
                # print("NO TIMESTAMP SUPPORT")
            else:
                if self.ts_fixed:
                    # print("Fixed timestamp")
                    max_payload_size_first = max_payload_size - self.ts_size
                else:
                    # print("timestamp variable.")
                    self.selectable_timestamps = True

        except Exception as e:
            raise TypeError(f"DAQ_INFO corrupted: {e}") from e

        # DAQ optimization.
        for daq_list in self.daq_lists:
            if self.selectable_timestamps:
                if daq_list.enable_timestamps:
                    max_payload_size_first = max_payload_size - self.ts_size
                else:
                    max_payload_size_first = max_payload_size
            ttt = make_continuous_blocks(daq_list.measurements, max_payload_size, max_payload_size_first)
            daq_list.measurements_opt = first_fit_decreasing(ttt, max_payload_size, max_payload_size_first)
        byte_order = 0 if self.xcp_master.slaveProperties.byteOrder == "INTEL" else 1
        self._first_pids = []
        daq_count = len(self.daq_lists)
        self.xcp_master.freeDaq()

        # Allocate
        self.xcp_master.allocDaq(daq_count)
        measurement_list = []
        for i, daq_list in enumerate(self.daq_lists, self.min_daq):
            measurements = daq_list.measurements_opt
            measurement_list.append((i, measurements))
            odt_count = len(measurements)
            self.xcp_master.allocOdt(i, odt_count)
        # Iterate again over ODT entries -- we need to respect sequencing requirements.
        for i, measurements in measurement_list:
            for j, measurement in enumerate(measurements):
                entry_count = len(measurement.entries)
                self.xcp_master.allocOdtEntry(i, j, entry_count)
        # Write DAQs
        for i, daq_list in enumerate(self.daq_lists, self.min_daq):
            measurements = daq_list.measurements_opt
            for j, measurement in enumerate(measurements):
                if len(measurement.entries) == 0:
                    continue  # CAN special case: No room for data in first ODT.
                self.xcp_master.setDaqPtr(i, j, 0)
                for entry in measurement.entries:
                    self.xcp_master.writeDaq(0xFF, entry.length, entry.ext, entry.address)

        # arm DAQ lists -- this is technically a function on its own.
        for i, daq_list in enumerate(self.daq_lists, self.min_daq):
            # print(daq_list.name, daq_list.event_num, daq_list.stim)
            mode = 0x00
            if self.supports_timestampes and (self.ts_fixed or (self.selectable_timestamps and daq_list.enable_timestamps)):
                mode = 0x10
            if daq_list.stim:
                mode |= 0x02
            ###
            ## mode |= 0x20
            ###
            self.xcp_master.setDaqListMode(
                daqListNumber=i, mode=mode, eventChannelNumber=daq_list.event_num, prescaler=1, priority=0xFF
            )
            res = self.xcp_master.startStopDaqList(0x02, i)
            self._first_pids.append(res.firstPid)
        if start_datetime:
            pass
        self.measurement_params = MeasurementParameters(
            byte_order,
            header_len,
            self.supports_timestampes,
            self.ts_fixed,
            self.supports_prescaler,
            self.selectable_timestamps,
            self.ts_scale_factor,
            self.ts_size,
            self.min_daq,
            self.start_datetime,
            self.daq_lists,
            self._first_pids,
        )
        self.set_parameters(self.measurement_params)

    def start(self):
        self.xcp_master.startStopSynch(0x01)

    def stop(self):
        self.xcp_master.startStopSynch(0x00)

    def first_pids(self):
        return self._first_pids


class DaqRecorder(DaqProcessor, _DaqRecorderPolicy):

    def __init__(self, daq_lists: List[DaqList], file_name: str, prealloc: int = 200, chunk_size: int = 1):
        DaqProcessor.__init__(self, daq_lists)
        _DaqRecorderPolicy.__init__(self)
        self.file_name = file_name
        self.prealloc = prealloc
        self.chunk_size = chunk_size

    def initialize(self):
        metadata = self.measurement_params.dumps()
        _DaqRecorderPolicy.create_writer(self, self.file_name, self.prealloc, self.chunk_size, metadata)
        _DaqRecorderPolicy.initialize(self)

    def finalize(self):
        _DaqRecorderPolicy.finalize(self)

    def start(self):
        DaqProcessor.start(self)


class DaqOnlinePolicy(DaqProcessor, _DaqOnlinePolicy):
    """Base class for on-line measurements.
    Handles multiple inheritence.
    """

    def __init__(self, daq_lists: List[DaqList]):
        DaqProcessor.__init__(self, daq_lists)
        _DaqOnlinePolicy.__init__(self)

    def start(self):
        DaqProcessor.start(self)


class DaqToCsv(DaqOnlinePolicy):
    """Save a measurement as CSV files (one per DAQ-list)."""

    def initialize(self):
        self.log.debug("DaqCsv::Initialize()")
        self.files: Dict[int, TextIO] = {}
        for num, daq_list in enumerate(self.daq_lists):
            if daq_list.stim:
                continue
            out_file = open(f"{daq_list.name}.csv", "w")
            self.files[num] = out_file
            hdr = ",".join(["timestamp0", "timestamp1"] + [h[0] for h in daq_list.headers])
            out_file.write(f"{hdr}\n")

    def on_daq_list(self, daq_list: int, ts0: int, ts1: int, payload: list):
        self.files[daq_list].write(f"{ts0},{ts1},{', '.join([str(x) for x in payload])}\n")

    def finalize(self):
        self.log.debug("DaqCsv::finalize()")
        ##
        ## NOTE: `finalize` is guaranteed to be called, but `Initialize` may fail for reasons.
        ##       So if you allocate resources in `Initialize` check if this really happened.
        ##
        if hasattr(self, "files"):
            for f in self.files.values():
                f.close()



#!/usr/bin/env python
"""Lowlevel API reflecting available XCP services

.. note:: For technical reasons the API is split into two parts;
          common methods and a Python version specific part.

.. [1] XCP Specification, Part 2 - Protocol Layer Specification
"""
from .master import Master  # noqa: F401
